"""
Internal message dispatcher
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


from threading import Thread, Lock
from Queue import Queue
from cxutil.UID import UID
from cxnet import zeroconf

import types
import traceback

from cxcore.cxnetlink import cl_socket, cl_lockd, mDNSengine, CL_DOMAIN, CL_PORT

def skip(*opts):
	pass

def dump(packet):
	print " ---- "
	print "\tto: %s" % (packet["to"])
	print "\tfrom: %s" % (packet["from"])
	print "\tdata: ",
	print packet["data"]


class connector(object):
	"""
	Two-way thread-safe bus connector for message delivering between two objects.
	It can not be used for any dispatcher-specific operations like subscribing etc, only
	for message transporting.
	
	Attributes:
		* rx (Queue) -- receive queue
		* tx (Queue) -- transmit queue
		* address (UID) -- unical identifier that is used by dispatcher2 to find a recipient
	
	Usually, you don't need to create an instances of this class by hands, it is automatically
	created by dispatcher2.connect()
	
	TODO:
	* create close() and check() methods
	* use weak references to already created Queue() instances
	* define and raise own exceptions
	"""
	
	rx = None
	tx = None
	address = None
	args = None
	
	def __init__(self,wires=(None,None)):
		'''
		Creates new connector for virtual "path cord"  with two wires ("full duplex" :)
		
		It takes a tuple as argument. The first member is a "tx" wire, the second -- "rx".
		If rx or tx "wire" is None, then it will be created.
		'''
		if wires[0]: self.tx = wires[0]
		else: self.tx = Queue()
		if wires[1]: self.rx = wires[1]
		else: self.rx = Queue()
		self.address = UID()
		# recursive reference to help logging
		self.bus = self

	def pre(self,args):
		'''
		Pre-input hook
		'''
		pass

	def cross(self):
		'''
		Returns a tuple of two wires that can be used to create the second connector
		for this "path cord". Note, that wires are crossed over in the tuple.
		'''
		return (self.rx,self.tx)

	def empty(self):
		'''
		Returns True, if there are no packets in rx
		'''
		return self.rx.empty()

	def drop(self):
		'''
		Drop packets from rx queue
		'''
		try:
			while True:
				self.rx.get_nowait()
		except:
			pass

	def put(self,address,message):
		'''
		Create a packet with "to" and "from" address fields and put it into the tx wire.
		Parameters:
			address (string) -- an address of a recipient
			message (any) -- a message that must be delivered
		
		If address is None, the message must be transported as is (connector assumes
		that it is already a packet)
		'''
		self.pre(self.args)

		if address:
			packet = {}
			packet["to"] = address
			packet["from"] = self.address
			packet["data"] = message
		else:
			packet = message
		result = self.tx.put(packet)
		return result

	def get(self,addr={},timeout=None):
		'''
		Get a packet from rx wire, strip data from it and return
		'''
		try:
			packet = self.rx.get(True,timeout)
		except:
			return None
		addr["to"] = packet["to"]
		addr["from"] = packet["from"]
		return packet["data"]
	
	def put2(self,address,message,addr={}):
		'''
		Put a message in tx and the get from rx
		'''
		self.put(address,message)
		return self.get(addr)

	def get_nowait(self,addr={}):
		'''
		Non-blocking get from rx wire
		'''
		try:
			packet = self.rx.get_nowait()
			addr["to"] = packet["to"]
			addr["from"] = packet["from"]
			result = packet["data"]
		except:
			result = False
		
		return result

class dispatcher2(Thread):
	"""
	Multiple-in multiple-out thread-safe message dispatcher
	
	It works like a router engine. It keeps two major dictionaries:
		* __arp_ -- a hash table, where by UID of client's connector you can find local connector
		* __multicast_ -- a dictionary, where by name of multicast group you can find a list of
		  local connectors of subscribed lines
	
	A message, with which the dispatcher2 works, is a dictionary with three fields.
		* "to" -- UID of a recipient or name of a group
		* "from" -- UID of a sender connector
		* "data" -- the data that must be transported
	
	When a message arrives, dispatcher2 looks into "to" field of the message and chooses a group
	or a particular recipient which must receive this message. A group is created when the first
	client calls subscribe() method.
	
	TODO:
	* create and raise own exeptions like "destination unreachable" etc.
	"""

	__input_ = None
	__multicast_ = None
	__arp_ = None
	## FIXME: make locking while any calls!
	## FIXME: vrotmnenogi, why?
	__subscribe_lock = None

	# routing engine
	server = None
	sock = None
	
	def __init__(self,name="dispatcher2",sname=None,address="0.0.0.0",join=[],port=CL_PORT,psk=None):
		'''
		Creates a new dispatcher2 thread named by name argument
		
		name		- thread name
		sname		- the service name (for mDNS registration)
		address		- address to bind the service on
		port		- port to bind the service on
		join		- string or tuple of strings - address[es] to use for set up multicast membership
		'''
		Thread.__init__(self)
		self.setName(name)
		self.__input_ = Queue()
		self.__arp_ = {}
		self.__multicast_ = {}
		self.hook = skip

		if sname:
			# set up ZeroConf mDNS
			self.sname = sname
			self.mdns = mDNSengine(address, join, self.connect())
			self.mdns.registerDomain(CL_DOMAIN)
			self.svc = self.mdns.registerService(sname,address,port,type="%s." % (CL_DOMAIN),properties={"state":"running","role":"_cx._udp. testbed at %s" % (sname)})

			# set up lockd
			self.lockd = cl_lockd(self.connect(), self.mdns, "%s.%s." % (sname,CL_DOMAIN))
			self.lockd.start()
			self.subscribe(self.lockd.bus.address, "lockd")

			# set up routing
			self.sock = cl_socket(address,port,self.connect(),keyfile=psk,debug=False)


	def cl_recv(self):
		self.__input_.put(self.sock.recv())


	def cl_route(self, packet):
		"""
		Analyzes a packet "to" address
		1. if it is targeted to this instance, strips host part
		2. if it is targeted outside, sends it to the target
		"""
		#
		# FIXME: use routing cache, do not split() and lookup
		# an address every time
		#
		t = packet["to"].split("@")
		# no host part specified
		if len(t) > 1:
			# t[0] -- subsystem address
			# t[1] -- host name
			# t[>1] -- just ignore
			
			# FIXME: see, why there is a bug when using such condition:
			# if t[1] in (self.sname, "%s.%s." % (self.sname, CL_DOMAIN)):

			if t[1] == self.sname:
				packet["to"] = t[0]
				return False

			try:
				host = self.mdns.lookup(t[1])[0]
				# lookup() returns ('ip.addr',port)
				packet["from"] = "%s@%s.%s." % (packet["from"],self.sname,CL_DOMAIN)
				self.sock.send(host,packet)
				return True
			except Exception,e:
				self.put({"to":"logger", "from":"dispatcher2", "data":("info","dispatcher2.cl_route(): %s (for %s)" % (str(e),t))})
				self.put({"to":"logger", "from":"dispatcher2", "data":("debug","%s" % (traceback.format_exc()) )})
				pass
		
		return False

	def shutdown(self):
		'''
		Shutdown cleanup
		'''
		self.put({"to":"logger", "from":"dispatcher2", "data":("info","dispatcher thread " + self.getName() + " shutdown")})
		self.broadcast("shutdown")
		if self.server:
			self.server.close()


	def connect(self):
		"""
		Returns a client's connector for new path cord
		"""
		## create the local connector
		local = connector((None,self.__input_))
		## cross-over client's connector
		client = connector(local.cross())
		## register local for client's uid
		self.__arp_[client.address] = local
		return client

	def disconnect(self,address):
		'''
		Remove a client from arp
		'''
		if self.__arp_.has_key(address):
			del self.__arp_[address]
			return True
		else:
			return False
	
	def subscribe(self,address,group):
		'''
		Subscribes an address to a multicast group, creates a group if there is no
		such key
		'''
		if not self.__multicast_.has_key(group):
			self.__multicast_[group] = []
		
		self.__multicast_[group].append(self.__arp_[address])

	def unsubscribe(self,address,group):
		'''
		Remove a subsciption from a multicast group
		'''
		if self.__multicast_.has_key(group):
			if self.__arp_[address] in self.__multicast_[group]:
				self.__multicast_[group].remove(self.__arp_[address])

	def run(self):
		'''
		Main retranslation cycle
		'''
		self.put({"to":"logger","from":"dispatcher2","data":("info","dispatcher thread " + self.getName() + " startup")})
		while True:
			# get a message
			packet = self.__input_.get()
			self.hook(packet)

			# validate
			if not self.validate(packet):
				continue

			# do routing
			if self.cl_route(packet):
				continue

			# analyse it by itself
			if packet["to"] == self.getName():
				if self.magic(packet["data"],packet["from"]):
					return
				continue
			
			self.put(packet)

	def validate(self,message):
		'''
		Validate a packet
		'''
		if not isinstance(message,dict):
			return False

		if \
			(not message.has_key("to")) or \
			(not message.has_key("from")) or \
			(not message.has_key("data")):
			return False
		
		return True
		

	def broadcast(self,message):
		## process broadcast messages
		for i in self.__arp_.keys():
			self.__arp_[i].put(i,message)


	def put(self,packet):
		## search a group
		if self.__multicast_.has_key(packet["to"]):
			for i in self.__multicast_[packet["to"]]:
				i.put(None, packet)
		
		## search a unicast
		if self.__arp_.has_key(packet["to"]):
			self.__arp_[packet["to"]].put(None, packet)
		
	def magic(self,text,addr):
		'''
		Analyses text and tries to interpret it as a "magic" command. Returns True, if the dispatcher2
		instance must break run() cycle and exit thread
		'''

		if type(text) == types.StringType:
			if text == "shutdown":
				self.shutdown()
				return True

			elif text == "connect":
				self.put({"to": addr, "from": "dispatcher2", "data": self.connect()})

			elif text == "debug":
				self.hook = dump
				self.put({"to": addr, "from": "dispatcher2", "data": "packet debug has been turned ON"})

			elif text == "no debug":
				self.hook = skip
				self.put({"to": addr, "from": "dispatcher2", "data": "packet debug has been turned OFF"})
			
			elif text == "dns":
				if self.sock:
					text = "\n"
					name = ""
					# FIXME: it breaks incapsulation
					for i in self.mdns.server.cache.entries():
						if name != i.name:
							text += "\n%s\n" % (i.name)
							name = i.name
						text += "\t%s\t%s\t%s\n" % (zeroconf._CLASSES[i.clazz], zeroconf._TYPES[i.type], i)
					self.put({"to": addr, "from": "dispatcher2", "data": text})
				else:
					self.put({"to": addr, "from": "dispatcher2", "data": "mDNS service is disabled"})

			elif text[:5] == "state":
				if self.sock:
					self.svc.setProperty("state",text.split()[1])

			elif text == "mdns":
				if self.sock:
					self.put({"to": addr, "from": "dispatcher2", "data": self.mdns})
				else:
					self.put({"to": addr, "from": "dispatcher2", "data": None})

			elif text[:3] == "ptr":
				if self.sock:
					try:
						result = "%s:\n" % (text.split()[1])
						for i in self.mdns.lookupPTR(text.split()[1]):
							result += "\t%s\n" % (i)
					except:
						result = "ptr error: %s" % (traceback.format_exc())
					
					self.put({"to": addr, "from": "dispatcher2", "data": "%s" % (result)})
				else:
					self.put({"to": addr, "from": "dispatcher2", "data": "mDNS service is disabled"})

			elif text[:6] == "lookup":
				if self.sock:
					try:
						host = self.mdns.lookup(text.split()[1],types=[])
						text = "address: %s:%s\nproperties:\n" % (host[0][0],host[0][1])
						for (i,k) in host[1].items():
							text += "\t%s:\t%s\n" % (i,k)
					except:
						text = "lookup error: %s" % (traceback.format_exc())
					
					self.put({"to": addr, "from": "dispatcher2", "data": "%s" % (text)})
				else:
					self.put({"to": addr, "from": "dispatcher2", "data": "mDNS service is disabled"})

			elif (text == "register_socket") and self.sock:
				self.put({"to": "monitor", "from": "dispatcher2", "data": ("register", self.sock.fd, self.cl_recv)})

		elif type(text) == types.DictType:
			if text.has_key("disconnect"):
				self.disconnect(text["disconnect"])

			if text.has_key("subscribe"):
				self.subscribe(text["subscribe"],text["group"])

			if text.has_key("hook"):
				# set hook for a connection
				c = self.__arp_[addr]
				c.pre = text["hook"]
				c.args = text["args"]

		## no reasons to break the main cycle found so far, return False...
		return False

	def arp(self):
		return self.__arp_
	
	def groups(self):
		return self.__multicast_

	def subscribers(self,group):
		return self.__multicast_[group]
