"""
This module implements a simple XML-RPC transport over AF_UNIX socket.
The reason is that we can obtain pid/uid/gid of a connected process with
simple getsockopt() call. So, we can delegate user authentication to the OS
level.

An operator can use standard terminal access via ssh or telnet. Our console
client must be used as a login shell. Once it has started, all authorization should
be done with our authorization core.
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from socket import SOL_SOCKET
import struct
import xmlrpclib
import traceback

from SimpleXMLRPCServer import SimpleXMLRPCRequestHandler, SimpleXMLRPCDispatcher
from SocketServer import UnixStreamServer, TCPServer
TCPServer.allow_reuse_address = True

from mangle import mangler
from threading import Thread, Lock
from cxutil.UID import UID
from cxcore.manager import ManagerThread

from executive.runner import runner as _runner
import orb.auth

import os
import sys
import pwd
import grp

import types

if sys.platform[:5] == "linux":
	SO_PEERCRED = 17
elif sys.platform[:7] == "freebsd":
	from ctypes.util import find_library
	from ctypes import c_uint32,CDLL,byref
	libc = CDLL(find_library("c"))
	

class DebugXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
	'''
	Wrapper for SimpleXMLRPCRequestHandler. Gets credentials info
	'''

	def do_POST(self):
		try:
			data = self.rfile.read(int(self.headers["content-length"]))
			response = self.server._marshaled_dispatch(data, getattr(self, '_dispatch', None))

			if self.protodebug:
				print data
				print response

		except:
			self.send_response(500)
			self.end_headers()
			raise
		else:
			# got a valid XML RPC response
			self.send_response(200)
			self.send_header("Content-type", "text/xml")
			self.send_header("Content-length", str(len(response)))
			self.end_headers()
			self.wfile.write(response)
		
			# shut down the connection
			self.wfile.flush()
			self.connection.shutdown(1)

class UnixXMLRPCRequestHandler(SimpleXMLRPCRequestHandler):
	'''
	Wrapper for SimpleXMLRPCRequestHandler. Gets credentials info
	'''

	def do_POST(self):
		try:
			## get peer credentials
			if sys.platform[:5] == "linux":
				(pid,uid,gid) = struct.unpack("III",self.rfile._sock.getsockopt(SOL_SOCKET,SO_PEERCRED,12))
			elif sys.platform[:7] == "freebsd":
				pid = 0
				uid = c_uint32()
				gid = c_uint32()
				a = libc.getpeereid(self.rfile._sock.fileno(), byref(uid), byref(gid))
				uid = uid.value
				gid = gid.value
			else:
				raise

			data = self.rfile.read(int(self.headers["content-length"]))

			## inject credentials into params and repack back into XML-RPC request
			## 8<------------------------------------
			params, method = xmlrpclib.loads(data)
			params = list(params)
			params[0]["pid"] = pid
			params[0]["user"] = pwd.getpwuid(uid).pw_name
			params[0]["group"] = grp.getgrgid(gid).gr_name
			data = xmlrpclib.dumps(tuple(params),method)
			## 8<------------------------------------
			

			response = self.server._marshaled_dispatch(data, getattr(self, '_dispatch', None))

			if self.protodebug:
				print data
				print response

		except:
			self.send_response(500)
			self.end_headers()
			raise
		else:
			# got a valid XML RPC response
			self.send_response(200)
			self.send_header("Content-type", "text/xml")
			self.send_header("Content-length", str(len(response)))
			self.end_headers()
			self.wfile.write(response)
		
			# shut down the connection
			self.wfile.flush()
			self.connection.shutdown(1)

class DebugXMLRPCServer(TCPServer, SimpleXMLRPCDispatcher):
	'''
	Simple XML-RPC server, implemented over AF_TCP sockets
	'''
	
	def __init__(self, addr, requestHandler=DebugXMLRPCRequestHandler, logRequests=0, debug=False):
		self.logRequests = logRequests
		if sys.version_info[1] == 4:
			# python 2.4
			SimpleXMLRPCDispatcher.__init__(self)
		elif sys.version_info[1] == 5:
			# python 2.5
			SimpleXMLRPCDispatcher.__init__(self,allow_none = False, encoding = "utf-8")
		else:
			# any other version
			raise Exception("Unsupported python version")
		requestHandler.protodebug = debug
		TCPServer.__init__(self, addr, requestHandler)

class UnixXMLRPCServer(UnixStreamServer, SimpleXMLRPCDispatcher):
	'''
	Simple XML-RPC server, implemented over AF_UNIX sockets
	'''
	
	def __init__(self, addr, requestHandler=UnixXMLRPCRequestHandler, logRequests=0, trusted=False, debug=False):
		self.logRequests = logRequests
		# FIXME: check version
		if sys.version_info[1] == 4:
			# python 2.4
			SimpleXMLRPCDispatcher.__init__(self)
		elif sys.version_info[1] == 5:
			# python 2.5
			SimpleXMLRPCDispatcher.__init__(self,allow_none = False, encoding = "utf-8")
		else:
			# any other version
			raise Exception("Unsupported python version")

		if trusted:
			requestHandler = DebugXMLRPCRequestHandler
		requestHandler.protodebug = debug
		UnixStreamServer.__init__(self, addr, requestHandler)


class register(object):
	'''
	Registration manager
	'''
	bus = None

	def __init__(self,bus,prefixes):
		self.bus = bus

	def register_service(self,credentials,sid,path):
		'''
		Register an external service
		'''
		return self.bus.put2("orb",
			{
				"map": "register_service",
				"id": sid,
				"path": path
			}
		)

	def register_module(self,credentials,sid,path):
		'''
		Register a module
		'''
		xInstance = orb.auth.proxy(path)
		if xInstance:
			return self.bus.put2("orb",
				{
					"map": "register_module",
					"sid": sid,
					"xClass": orb.auth.proxy,
					"xInstance": xInstance
				}
			)
		else:
			return False

	def register_acl(self,credentials,path,mode,name):
		'''
		Register ACL entry
		'''
		return self.bus.put2("orb",
			{
				"map": "register_acl",
				"mode": mode,
				"name": name,
				"path": path,
			}
		)

	def unregister_acl(self,credentials,path,mode,name):
		'''
		Unregister ACL entry
		'''
		return self.bus.put2("orb",
			{
				"map": "unregister_acl",
				"mode": mode,
				"name": name,
				"path": path,
			}
		)


class sessionManager(ManagerThread):
	'''
	Session manager
	'''
	def __init__(self,bus,prefixes):
		ManagerThread.__init__(self,bus,"session manager thread")

		self.prefixes = prefixes
		self.sessions = {}

	def init(self,opts):
		return {"prompt": self.prefixes[0]}

	def _start(self,opts):
		user = opts["credentials"]["user"]
		sid = opts["sid"]
		credentials  = opts["credentials"]
		
		if not self.sessions.has_key(user):
			self.sessions[user] = {}

		if not self.sessions[user].has_key(sid):
			self.sessions[user][sid] = session(
				bus = self.bus.put2("dispatcher2","connect"),
				sid = sid,
				credentials = credentials,
				prefixes = self.prefixes
			)
		return sid

	def check(self,opts):
		user = opts["credentials"]["user"]
		sid = opts["sid"]

		if self.sessions.has_key(user):
			if self.sessions[user].has_key(sid):
				return True

		return False

	def parse(self,opts):
		credentials = opts["credentials"]
		user = credentials["user"]
		sid = opts["sid"]
		commands = opts["commands"]
		cookie = opts["cookie"]
		noout = opts["noout"]

		if not self.check({"credentials":credentials,"sid":sid}):
			return self.error("no such session")

		for i in commands:
			# FIXME:
			# That's a bug: if we gather result in cycle, it MUST be a list;
			# at least, we MUST initialize it
			result = self.sessions[user][sid].parse(credentials, cookie,i[0].encode('utf-8'),i[1],noout)

		return result

	def interrupt(self,opts):
		credentials = opts["credentials"]
		user = credentials["user"]
		sid = opts["sid"]
		code = opts["code"]
		
		if not self.check({"credentials":credentials,"sid":sid}):
			return self.error("no such session")

		return self.sessions[user][sid].interrupt(credentials,code)

	def input(self,opts):
		credentials = opts["credentials"]
		user = credentials["user"]
		sid = opts["sid"]
		text = opts["text"]

		if not self.check({"credentials":credentials,"sid":sid}):
			return self.error("no such session")

		return self.sessions[user][sid].input(credentials,text.encode('utf-8'))

	def terminate(self,opts):
		credentials = opts["credentials"]
		user = credentials["user"]
		sid = opts["sid"]

		if not self.check({"credentials":credentials,"sid":sid}):
			return False

		self.sessions[user][sid].terminate()
		del self.sessions[user][sid]

		return True

	def error(self,report):
		return {"error": report}

class session(object):
	'''
	Session object: must be destroyed and distributed between runner and mangler objects
	'''
	def __init__(self,bus,sid,credentials,prefixes):
		

		self.sid = sid
		self.prefixes = prefixes
		self.bus = bus
		self.mangler_bus = bus.put2("dispatcher2","connect")
		self.runner_bus = bus.put2("dispatcher2","connect")

		self.runner = _runner(self.runner_bus,"runner thread for " + str(sid))
		self.mangler = mangler(self.mangler_bus,self.runner.bus.address,sid,self.prefixes)
		self.runner.start()

	def parse(self,credentials,cookie,command,lines,noout):
		try:
			self.mangler.queue += 1
			return self.mangler.parse(credentials,cookie,command,lines,noout)
		except Exception,e:
			print e
			traceback.print_exc()

	def interrupt(self,credentials,code):
		return self.mangler.interrupt(credentials,code)

	def input(self,credentials,text):
		try:
			return self.mangler.input(credentials,text)
		except Exception,e:
			print e
			traceback.print_exc()

	def terminate(self):

		self.bus.put(self.runner.bus.address,"shutdown")

		# FIXME: each object MUST start and shutdown connections by ITSELF
		self.bus.put("dispatcher2",{"disconnect": self.runner.bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.runner.orb_bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.mangler.bus.address})
		self.bus.put("dispatcher2",{"disconnect": self.mangler.busInstance.address})
		self.bus.put("dispatcher2",{"disconnect": self.mangler.busRunner.address})
		self.bus.put("dispatcher2",{"disconnect": self.mangler.busState.address})
		
		os.close(self.mangler.pipe_read)
		os.close(self.mangler.pipe_write)
		os.close(self.mangler.ctl_read)
		os.close(self.mangler.ctl_write)

class proxy(object):
	'''
	A session manager that creates and maintains mangler instances while a session is running.
	'''
	bus = None
	
	def __init__(self,bus):
		self.bus = bus
	def check(self,credentials,sid):
		return self.bus.put2("session manager", {"map":"check","credentials":credentials,"sid":sid})
	def init(self,credentials,opts = {}):
		return self.bus.put2("session manager", {"map":"init"})
	def start(self,credentials):
		return self.bus.put2("session manager", {"map":"_start","credentials":credentials,"sid":UID()})
	def parse(self,credentials,cookie,sid,commands,noout=False):
		return self.bus.put2("session manager",{"map":"parse","credentials":credentials,"cookie":cookie,"sid":sid,"commands":commands,"noout":noout})
	def interrupt(self,credentials,sid,code):
		return self.bus.put2("session manager",{"map":"interrupt","credentials":credentials,"sid":sid,"code":code})
	def input(self,credentials,sid,text):
		return self.bus.put2("session manager",{"map":"input","credentials":credentials,"sid":sid,"text":text})
	def terminate(self,credentials,sid):
		return self.bus.put2("session manager",{"map":"terminate","credentials":credentials,"sid":sid})

class server(Thread):
	'''
	XML-RPC publisher
	'''
	flags = None
	server = None
	bus = None

	def __init__(self,bus,publish,name="XML-RPC server",socket="socket",trusted=False,debug=False):
		Thread.__init__(self)
		self.setName(name)
		self.flags = {}
		self.bus = bus
		
		try:
			if type(socket) == types.TupleType:
				self.server = DebugXMLRPCServer(socket, debug = debug)
			else:
				self.server = UnixXMLRPCServer(socket, trusted = trusted, debug = debug)
			self.server.register_introspection_functions()
			self.server.register_instance(publish(self.bus))
		except:
			raise

	
	def run(self):
		while True:
			self.server.handle_request()
			if self.bus.get_nowait() == "shutdown":
				return

