"""
Various internal-use classes:

 * authorization facility
 * taboo
 * proxy for external modules

"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

from popen2 import Popen3
import os
from command import command, swrapper
import types
from new import classobj

from bus import transport

import pwd
import grp

class auth(object):
	"""
	Authorization proxy. It has nothing to do with authentication: there is
	no password checking or like that. An auth object believes that the
	username supplied is correct. In other words, an auth object trusts
	the application core and supposes users authenticated in any way.
	
	The only goal of an auth object is to permit or deny access to a module
	path for a user. For this, an auth object maintains internal dictionary
	of acl. If there is a key in the dictionary for given module path, the object
	looks, is there a policy for a username supplied. If there is no such policy,
	it uses a default policy for the path, if there is no default policy or no such
	key in the dictionary, it uses global default policy. If there is no default
	policy at all, it permits access for 'root' and denies access for other users.
	"""
	acl = None
	
	def __init__(self):

		

		self.acl = {
			"user": {
				'*': [pwd.getpwuid(os.getuid()).pw_name,],
			},
			"group": {
				'*': [grp.getgrgid(os.getgid()).gr_name,],
			}
		}

	def check_direct(self,name,mode,url):
		if self.acl[mode].has_key(url):
			if name in self.acl[mode][url]:
				return True
		return False

	def check_asterisk(self,name,mode,url):
		if self.acl[mode].has_key("*"):
			if name in self.acl[mode]["*"]:
				return True

		for (i,k) in self.acl[mode].items():
			if i[-1] == "*":
				i = i[:-1]
				if (url.startswith(i)) and (name in k):
					return True

		return False
	
	def check(self,name,mode,url):
		if self.check_direct(name,mode,url):
			return True
		else:
			return self.check_asterisk(name,mode,url)

	def register(self,name,mode,url):
		if not self.check_direct(name,mode,url):
			if not self.acl[mode].has_key(url):
				self.acl[mode][url] = []
			self.acl[mode][url].append(name)
		return True

	def unregister(self,name,mode,url):
		if self.check_direct(name,mode,url):
			self.acl[mode][url].remove(name)
			if not self.acl[mode][url]:
				del self.acl[mode][url]
			return True
		return False

class taboo(command):
	'''
	An instance of this class must be returned if permission is not granted to an object
	'''
	def postUp(self,opts):
		result = swrapper()
		result.add_commands("message", ["Permission denied\n",])
		return result

class proxy(command):
	'''
	Proxy object for an external module
	'''
	path = None
	pheader = None
	pfooter = None
	capabilities = None
	default_service = None
	
	def __init__(self,path):
		'''
		Init self.parameters from the module
		'''

		if path == "empty":
			command.__init__(self)
			self.begin = True
			self.bypass = False
			return None

		self.path = path
		if not os.access(self.path,os.X_OK):
			self.broken = True
			return None

		command.__init__(self)

		self.default_service = "message"
		self.capabilities = {
			"services": False,
		}

		message = self.action((("action","description"),),analyze=True)
		message = message.splitlines()
		for i in message:
			i = i.split(":")
			if len(i) < 4:
				continue

			xName = i[0].strip()
			xType = i[1].strip()
			xList = i[2].strip()
			if xList == "false":
				xList = False
			else:
				xList = True

			xDesc = i[3].strip()
			xClass = classobj(xName,(command,),{})
			xClass.doc = xDesc
			xClass.parameters = {
				"$1": {
					"name": xName,
					"type": xType,
					"list": xList,
					"description": xDesc
				}
			}
			xClass.modules = []
			if not xList:
				xClass.unique = True

			setattr(self,xName,xClass)
			self.modules.append(xName)

		if len(self.modules) > 0:
			self.begin = True
			self.bypass = False
		
		capabilities = self.action((("action","capabilities"),),analyze=True)
		capabilities = capabilities.splitlines()
		for i in capabilities:
			i = i.split(":")
			if len(i) < 2:
				continue

			xName = i[0].strip()
			xValue = i[1].strip()
			if xName == "capability":
				self.capabilities[xValue] = True
			elif xName == "default-service":
				self.default_service = xValue

	def action(self,actions,analyze=False):
		'''
		'''
		inst = Popen3([self.path],True,-1)
		(i,k,l) = (inst.fromchild, inst.tochild, inst.childerr)

		messages = []
		for (x,y) in actions:
			messages.append(transport.encode("%s\n%s" % (x,y)))

		k.write(transport.multipart(messages).as_string())
		k.write("\n")
		k.close()
		message = i.read()
		status = os.waitpid(inst.pid,0)
		retval = 255
		if os.WIFEXITED(status[1]):
			retval = os.WEXITSTATUS(status[1])

		if retval != 0 and analyze:
			self.broken = True
			return None

		return message

	def postUp(self,opts):
		'''
		"write" action
		'''
		return self.post(opts,"write")

	def postDown(self,opts):
		'''
		"delete" action
		'''
		return self.post(opts,"delete")

	def post(self,opts,action):
		'''
		Pass options to the module
		'''
		
		message = []

		for param in opts["envl"].keys():
			if type(opts["envl"][param]) == types.ListType:
				for value in opts["envl"][param]:
					message.append([param,value])
					message.append(["action",action])
			else:
				message.append([param,opts["envl"][param]])
				message.append(["action",action])

		message.append(["type",action])
		message.append(["action","commit"])

		message = self.action(message)
		
		result = swrapper()
		if self.capabilities["services"]:
			service = self.default_service
			message = message.splitlines()
			for i in message:
				if i.strip()[:2] == "#!":
					service = i.strip()[2:]
				else:
					result.add_commands(service,[i,])
		else:
			result.add_commands("message", [message, ])

		return result
