"""
State tree classes
"""

# 	Copyright (c) 2005-2008 Peter V. Saveliev
#
# 	This file is part of Connexion project.
#
# 	Connexion is free software; you can redistribute it and/or modify
# 	it under the terms of the GNU General Public License as published by
# 	the Free Software Foundation; either version 3 of the License, or
# 	(at your option) any later version.
#
# 	Connexion is distributed in the hope that it will be useful,
# 	but WITHOUT ANY WARRANTY; without even the implied warranty of
# 	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# 	GNU General Public License for more details.
#
# 	You should have received a copy of the GNU General Public License
# 	along with Connexion; if not, write to the Free Software
# 	Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import types
import inspect
import traceback

from copy import copy,deepcopy

from command import command,swrapper
from syntax import oparser,xSyntaxError,xSyntaxErrorNotCompleted,xSyntaxErrorNoMandatory,xParamError
from threading import Lock

from cxcore.manager import ManagerThread

from cxutil import utils
from cxutil import cmask
from cxutil import emask
from cxutil import flags
from cxutil.exceptions import CommitRaise,CallPass,CallQueue,BranchPass
from cxutil.UID import UID

SIGNAL_OID = UID()

def ssort(xI,yI):
	"""
	Compares two state tree nodes
	"""
	# compare node priorities
	if xI.module.priority == yI.module.priority:
		return utils.nsort(xI.fn, yI.fn)

	if xI.module.priority > yI.module.priority:
		return 1
	else:
		return -1

class state2(object):
	'''
	State tree node, version 2
	'''
	bus = None
	module = None		# node module
	fn = None		# node name + options
	cn = None		# node name
	opts = None
	comment = None

	creds = None
	cookie = None

	children = None
	parent = None
	modes = None

	globals = None
	privates = None
	persist = None

	condition = None
	transaction = None
	marked = None
	confirm = None
	executed = None

	functions = None
	filters = None
	declare = None
	declare_counter = None
	subscriptions = None

	def __init__(self, module=None, creds={"user": "nobody", "group": "nobody", "pid": 0}, cookie = "none", fn="", cn="", opts=oparser({},[]), parent=None, condition=None, bus=None):

		if module:
			self.module = module
		else:
			self.module = command()
			self.module.mask |= flags.Begin | flags.Bypass
			self.module.prefix = cn
			self.module.namespace = cn
			self.module.bypass = True

		self.creds = creds
		self.cookie = cookie
		self.fn = fn
		self.cn = cn
		self.opts = opts
		self.comment = []
		self.children = {}
		self.unique = {}
		self.parent = parent

		self.privates = utils.opts()
		self.persist = utils.opts()

		self.condition = condition
		self.transaction = {
			"up": [],
			"down": [],
		}
		self.marked = []
		self.confirm = []
		self.lines = []
		self.executed = False

		###
		# function calls support
		###
		self.declare_counter = 0
		self.declare = []

		###
		# public container for arbitrary attributes
		###
		self.public = utils.opts()

		if (not bus) and (self.parent):
			self.bus = self.parent.bus
		else:
			self.bus = bus

		self.module.bus = self.bus

		###
		# shared arrays and other dependants on the parent node
		###
		if self.parent:
			# shared
			self.globals = parent.globals
			self.subscriptions = self.parent.subscriptions
			self.catchers = self.parent.catchers
			self.modes = self.parent.modes
			self.functions = self.parent.functions
			self.scripts = self.parent.scripts
			self.filters = self.parent.filters
			self.queue = self.parent.queue
			# ...
			self.public.parent = self.parent.public
		else:
			# shared
			self.globals = utils.opts()
			self.subscriptions = {}
			self.catchers = {}
			self.functions = {}
			self.scripts = {}
			self.filters = {}
			self.modes = []
			self.queue = []
			# ...
			self.public.parent = None

	def pwdNS(self,ns=None,first=True):
		r = ''
		
		if first:
			first = False
		else:
			if ns:
				if self.module.namespace == ns:
					r = ' ' + self.cn
			else:
				ns = self.module.namespace
			
		if self.parent:
			return self.parent.pwdNS(ns,first) + r
		else:
			return ns

	def xmlDump(self,creds,cookie,oid):
		from xml.dom.minidom import getDOMImplementation
		imp = getDOMImplementation()
		dom = imp.createDocument(None, "node", None)
		node = dom.firstChild
		node.setAttribute("name",self.cn)

		# attach comment
		#k = dom.createElement("comment")
		#k.appendChild(dom.createTextNode(self.comment.__repr__()))
		#node.appendChild(k)
		for (i,k) in self.module.parameters.items():
			l = dom.createElement("parameter")
			node.appendChild(l)
			#
			# <parameter name="$2" rname="action" type="enum" mask="0">
			#	<desc>Rule action</desc>
			#	<value>masquerade</value>
			#	<range>
			#		<v>accept</v>
			#		<v>reject</v>
			#		<v>drop</v>
			#		<v>dnat</v>
			#		<v>snat</v>
			#		<v>masquerade</v>
			#		<v>set</v>
			#	</range>
			# </parameter>

			# required: set `name` attribute
			l.setAttribute("name",i)
			# optional: set `rname` attribute (when re-naming is needed, e.g. "$2" -> "action"
			if k.has_key("name"):
				l.setAttribute("rname",k["name"])
			# required: set type attribute
			l.setAttribute("type",k["type"])
			# required: set mask attribute
			l.setAttribute("mask",str(k["mask"]))

			# append description
			m = dom.createElement("desc")
			m.appendChild(dom.createTextNode(k["description"]))
			l.appendChild(m)

			# append value
			if self.opts.has_key(i):
				m = dom.createElement("value")
				m.appendChild(dom.createTextNode(str(self.opts[i])))
				l.appendChild(m)

			# append range (optional)
			if k.has_key("range"):
				m = dom.createElement("range")
				for n in k["range"]:
					o = dom.createElement("v")
					o.appendChild(dom.createTextNode(str(n)))
					m.appendChild(o)
				l.appendChild(m)

		# node.appendChild(dom.createTextNode(self.module.parameters.__repr__()))

		for (i,k) in self.children.items():
			node.appendChild(k.xmlDump(creds,cookie,oid).firstChild)
		return dom

	def pass_filter(self,data):
		return self.module.pass_filter(data)

	def pack(self,creds,cookie,module,opts,locs,shared,rest_fc,rest_tp,lines,oid,done=False,esoteric={},mask=0,comment=[],node=None):
		'''
		Create option set for a command object
		'''

		package = utils.opts({
				"creds": creds,
				"cookie": cookie,
				"opts": opts,
				"envg": self.globals,
				"envp": self.privates,
				"rest_fc": rest_fc,
				"rest_tp": rest_tp,
				"shared": shared,
				"persist": self.persist,
				"condition": self.condition,
				"lines": lines,
				"already": done,
				"fn": self.fn,
				"oid": oid,
				"envl": locs,
				"call_mask": copy(mask),
			})

		if node:
			package["public"] = node.public
		else:
			package["public"] = self.public

		if module.mask & flags.Esoteric:
			package["esoteric"] = esoteric
			if node:
				package["esoteric"]["node"] = node
			else:
				package["esoteric"]["node"] = self
			package["esoteric"]["root"] = self.getRoot()
			package["esoteric"]["comment"] = comment
			package["esoteric"]["call_mask"] = mask

		return package

	def call(self,creds,cookie,module,method,opts,locs,shared,rest_fc,rest_tp,lines,oid,done=False,esoteric={},mask=0,comment=[],node=None):
		'''
		Call an object and return a result message
		'''
		message = []
		result = None
		pack = self.pack(creds,cookie,module,opts,locs,shared,rest_fc,rest_tp,lines,oid,done,esoteric,mask,comment,node)
		try:
			result = method(pack)
		except CommitRaise,e:
			raise e
		except Exception,e:

			result = swrapper()
			text = "\n*** Error report begin ***\n"
			text += "error command: >%s<\n%s\nfailed: %s\n" % (self.fn,method,e)
			text +=	"----\n%s----\n" % (traceback.format_exc())
			text += "package was:\n"
			for i in pack.keys():
				if i == "esoteric":
					text += "\tesoteric:\n"
					for k in pack[i].keys():
						text += "\t\t%s: %s\n" % (k,pack[i][k])
						if type(pack[i][k]) == type(self):
							pfn = "<no parent>"
							if pack[i][k].parent:
								pfn = pack[i][k].parent.fn
							text += "\t\t  (fn: `%s`, parent: `%s`)\n" % (pack[i][k].fn, pfn)
				else:
					text += "\t%s: %s\n" % (i,pack[i])
			text += "***  Error report end  ***\n\n"
			result.add_commands("message",[text,])
			result.add_commands("logger",[text,])

		if type(result) == types.ListType:
			message.extend(result)
		else:
			if result:
				message.append(result)
		for i in message:
			if isinstance(i,swrapper):
				if i.empty():
					# clean up empty services
					message.remove(i)
					continue

				i.cookie = cookie
			else:
				self.bus.put("logger",("warning","deprecated type in module response: %s" % (type(i))))
		return message

	def abort(self,oid):
		'''
		Abort a transaction
		'''
		for (i,k) in self.children.items():
			k.abort()

		if not self.executed:
			self.clear()

	def cancel(self,oid,nested=False):
		'''
		Queue a node to abort, or abort immediately (if it was not executed)
		'''
		if self.executed:
			if not nested:
				self.mark(oid)
			self.transaction["down"].append(oid)
			for (i,k) in self.children.items():
				k.cancel(oid,True)
		else:
			self.clear()

	def clear(self):
		'''
		Remove a node itself and all references to it
		'''
		###
		# remove a node. Clean up:
		# 1. functions
		# 2. scripts
		# 3. unique
		# 4. subscriptions
		# 5. the children itself
		###
		# 1.
		for (k,l) in self.functions.items():
			if self == l:
				del self.functions[k]
		# 2.
		for (k,l) in self.scripts.items():
			if self in l:
				l.remove(self)
			if not l:
				del self.scripts[k]
		# 3.
		if self.unique.has_key(self.cn):
			del self.unique[self.cn]
		# 4.a
		for (k,l) in self.subscriptions.items():
			if self in l:
				l.remove(self)
			if not l:
				del self.subscriptions[k]
		# 4.b
		for (k,l) in self.catchers.items():
			if self in l:
				l.remove(self)
			if not l:
				del self.catchers[k]
		# 5.
		del self.parent.children[self.fn]

	def bypass(self):
		'''
		Returns True, if commands must be executed immediately, and False, if they must
		be queued.
		'''
		if not (self.module.mask & flags.Bypass):
			return False
		else:
			if self.parent:
				return self.parent.bypass()
			else:
				return True

	def parse(self,creds,cookie,opts,comment,oid,last,esoteric):

		message = []
		fn = opts[1] + opts[2]
		cn = opts[1]
		object = opts[0]
		parsed = None
		opts2 = opts[2].strip().split()
		lines = []
		done = False
		condition = None

		locs = utils.opts()
		shared = utils.opts()
		rest_fc = utils.opts()
		rest_tp = utils.opts()


		try:
			parsed = oparser(object.parameters,opts2)

		except Exception,e:
			s = swrapper()
			s.add_commands("message",["%s\n" % e,])
			return (None, [s,])


		if not object.mask & flags.Esoteric:
			esoteric = {}

		if self.children.has_key(fn):
			# confirm a node if `merge mode`
			if "merge" in self.modes:
				self.children[fn].confirm.append(oid)
			# ...
			# that can be only "up" transaction, so,
			# update transaction and marked arrays
			# if the node isn't executed yet
			if not self.children[fn].executed:
				if not oid in self.children[fn].transaction["up"]:
					self.children[fn].transaction["up"].append(oid)
				if not oid in self.children[fn].marked:
					self.children[fn].marked.append(oid)
			
			if self.children[fn].module.mask & flags.Begin:
				return (self.children[fn],message)
			else:
				return (None,[])

		# create a node
		node = state2(object,creds,cookie,fn,cn,parsed,self)

		# run immediate actions
		try:
			message.extend(self.call(creds, cookie, object, object.pre, parsed, locs, shared, rest_fc, rest_tp, lines, oid, esoteric=esoteric, mask=cmask.Immediate, comment=comment, node=node))
			message.extend(self.call(creds, cookie, object, object.immediate, parsed, locs, shared, rest_fc, rest_tp, lines, oid, esoteric=esoteric, mask=cmask.Immediate, comment=comment, node=node))
			if object.mask & flags.Begin:
				node.condition = locs
		except Exception,e:
			# just ignore all unwanted exceptions
			print e
			pass


		# choose, if we must run it instead of queueing it for commit
		if (object.mask & flags.Immediate) or (self.bypass() and not (object.mask & flags.Begin)):
			message.extend(self.call(creds, cookie, object, object.postUp, parsed, locs, shared, rest_fc, rest_tp, lines, oid, esoteric=esoteric, mask=cmask.Immediate, comment=comment, node=node))
			if last:
				node = None
			return (node,message)



		# check, if there is unique object
		if self.unique.has_key(cn):
			self.unique[cn].cancel(oid)
			del self.unique[cn]

		# if we have no such children yet, create it and mark tree up to the root
		self.children[fn] = node
		self.mark(oid)

		# confirm a node if `merge mode`
		if "merge" in self.modes:
			self.children[fn].confirm.append(oid)

		# attach a comment
		if comment:
			self.children[fn].comment = deepcopy(comment)

		# check, if the object is unique
		if object.mask & flags.Unique:
			self.unique[cn] = self.children[fn]

		# check subscriptions
		for i in object.subscribe:
			if not i in self.subscriptions:
				self.subscriptions[i] = []
			if not self.children[fn] in self.subscriptions[i]:
				self.subscriptions[i].append(self.children[fn])

		# check signal catchers
		for i in object.catch:
			if not i in self.catchers:
				self.catchers[i] = []
			if not self.children[fn] in self.catchers[i]:
				self.catchers[i].append(self.children[fn])

		# check for `transparent` attribute
		if object.mask & flags.Transparent:
			object.prefix = self.module.prefix

		# if the children isn't executed yet, add oid to the transaction
		if not self.children[fn].executed:
			self.children[fn].transaction["up"].append(oid)

		# analyze public container
		if node.public.has_key("script"):
			if not self.scripts.has_key(node.public.script):
				self.scripts[node.public.script] = []
			self.scripts[node.public.script].append(node)

		# analyze auto-children
		for (i,k) in object._children.items():
			(x,m) = self.parse(\
				creds=creds,\
				cookie=cookie,\
				opts=[k(),i.split()[-1],""],\
				comment="",\
				oid=oid,\
				last=False,\
				esoteric=esoteric)
			message.extend(m)

		if object.mask & flags.Begin:
			return (self.children[fn],message)
		else:
			return (self,message)

	def mark(self,oid,where="up"):
		'''
		'''
		if where == "up":
			if not oid in self.marked:
				self.marked.append(oid)
			if self.parent:
				self.parent.mark(oid)

		elif where == "down":
			if not oid in self.marked:
				self.marked.append(oid)
			for (i,k) in self.children.items():
				k.mark(oid,"down")

	def getRoot(self):
		if self.parent:
			return self.parent.getRoot()
		else:
			return self

	def fqcn(self):
		text = self.cn
		if self.parent:
			text = "%s %s" % (self.parent.fqcn(), text)
		return text

	def commit(self,creds,cookie,oid,l={},shared={},mask=0,esoteric={}):
		'''
		Commit transaction tree from this node and depthward
		'''

		up = False
		down = False
		message = []
		cmessage = []
		if not creds:
			creds = self.creds
		if not cookie:
			cookie = self.cookie

		###
		# Local variables, opts.envl. Are shared between this node and children.
		###
		locs = utils.opts(l)
		###
		# Variables, that children return to this node explicitly
		###
		rest_fc = utils.opts()
		###
		# Variables, that this node will return to parent
		###
		rest_tp = utils.opts()
		###
		# Split locals for postUp() and postDown() calls
		###
		locals_up = utils.opts()
		locals_down = utils.opts()
		###
		# Variables shared between children only
		###
		sd = utils.opts()


		result = {
			"services": [],
			"emit": [],
			"up": {},
			"down": {},
			"rest": {},
			"cancel": False,
		}

		if \
			( \
				(self.module.mask & flags.Begin) and\
				(not self.module.mask & flags.LocalsTransparent)\
			) or\
				(self.module.mask & flags.StopLocals):
			result["emit"].append("stopLocals")

		##
		# if the node is not in-service for the oid, skip sub-tree
		##
		if\
			(\
				(not oid in self.transaction["up"]) and\
				(not oid in self.transaction["down"]) and\
				(not oid in self.marked) and\
				(not mask & cmask.Force) and\
				(not "merge" in self.modes)\
			):
			return result

		if self.parent:
			if\
				(\
					self.parent.public.has_key("script") and\
					( (not mask & cmask.ScriptUp) and (not mask & cmask.ScriptDown) )\
				):
				mask |= cmask.Bypass


		if self.declare_counter > 0:
			if not mask & cmask.Call:
				mask |= cmask.Bypass
			mask |= cmask.Force

		if self.module.mask & flags.Force:
			mask |= cmask.Force

		## run pre-execution, if any
		if not mask & cmask.Bypass:
			try:
				message.extend(self.call(creds,cookie,self.module,self.module.pre,self.opts,locs,shared,rest_fc,rest_tp,self.lines,oid,self.executed,esoteric,mask))

			except BranchPass,e:
				# cmask.Force removed after false commit of unrelated
				# nodes under `catch *`; it is that after module.pre()
				# BranchPass is thrown and cmask.Force causes commit
				# not only transaction's nodes, but all children of
				# `catch *`
				mask |= cmask.Bypass # | cmask.Force
				result["emit"].append("up")

		## recursively commit transaction
		k = self.children.values()
		k.sort(ssort)
		for i in k:
			try:
				env = i.commit(creds,cookie,oid,locs,sd,mask,esoteric = esoteric)

				utils.merge(env["rest"],rest_fc)

				if ("up" in env["emit"]) and (not "stopLocals" in env["emit"]):
					utils.merge(locs,locals_up)
					utils.merge(env["up"],locals_up)
					up = True
				if ("down" in env["emit"]) and (not "stopLocals" in env["emit"]):
					utils.merge(locs,locals_down)
					utils.merge(env["down"],locals_down)
					down = True

				if env["cancel"]:
					i.clear()

				if env["services"]:
					cmessage.extend(env["services"])

			except BranchPass,e:
				# FIXME
				print "use BranchPass only in pre() calls"
				pass

			except CallPass,e:
				pass

			except CallQueue,e:
				if (self.module.mask & flags.Begin) and (not self.module.mask & flags.LocalsTransparent):
					self.queue.append(i)
				else:
					raise e

			except Exception,e:
				print "unknown exception: %s" % (e)
				print "----\n%s" % (traceback.format_exc())


		if (oid in self.transaction["down"]) or\
			(("merge" in self.modes) and (not oid in self.confirm)):
			result["cancel"] = True

		## choose methods to run
		if not mask & cmask.Bypass:

			if \
				(oid in self.transaction["down"]) or\
				(("merge" in self.modes) and (not oid in self.confirm)) or\
				(mask & cmask.ScriptDown):
				utils.merge(locs,locals_down)
				down = True

			elif mask & cmask.Force:
				utils.merge(locs,locals_down)
				utils.merge(locs,locals_up)
				down = up = True

			elif \
				(oid in self.transaction["up"]) or\
				(mask & cmask.ScriptUp):
				utils.merge(locs,locals_up)
				up = True

		else:
			if not self.executed:
				self.executed = True


		if up and (self.module.mask & flags.Newborn):
			self.module.mask = self.module.mask & ~flags.Newborn
			message.extend(self.call(creds,cookie,self.module,self.module.init,self.opts,locals_up,shared,rest_fc,rest_tp,self.lines,oid,self.executed,esoteric,mask))


		message.extend(cmessage)

		if down:
			message.extend(self.call(creds,cookie,self.module,self.module.postDown,self.opts,locals_down,shared,rest_fc,rest_tp,self.lines,oid,self.executed,esoteric,mask))
			if not "down" in result["emit"]:
				result["emit"].append("down")
			if result["cancel"]:
				message.extend(self.call(creds,cookie,self.module,self.module.terminate,self.opts,locals_down,shared,rest_fc,rest_tp,self.lines,oid,self.executed,esoteric,mask))

		if up:
			message.extend(self.call(creds,cookie,self.module,self.module.postUp,self.opts,locals_up,shared,rest_fc,rest_tp,self.lines,oid,self.executed,esoteric,mask))
			if not "up" in result["emit"]:
				result["emit"].append("up")


		## check subscriptions
		subs = self.subscribed()
		if subs:
			for i in subs:
				if i.declare_counter == 0:
					message.extend(i.commit(creds,cookie,oid,locs,{},cmask.Force | cmask.Call)["services"])

		result["services"] = message
		result["up"] = deepcopy(locals_up)
		result["down"] = deepcopy(locals_down)
		result["rest"] = deepcopy(rest_tp)

		if \
			(oid in self.confirm) and \
			(not mask & cmask.Call):
			self.confirm.remove(oid)

		self.purge(oid)
		self.privates = utils.opts()
		return result

	def subscribed(self):
		fqcn = self.fqcn()
		subs = []
		for i in self.subscriptions:
			if i[-1] == "*":
				k = i[:-1]
				if fqcn.startswith(k):
					subs.extend(self.subscriptions[i])
			else:
				if fqcn == i:
					subs.extend(self.subscriptions[i])
		return subs

	def purge(self,oid):
		# now called only from self.commit()
		
		###
		# FIXME: Razgrebi etu pomojku!
		###
		
		###
		#
		# completely clean up transaction, leaving only
		# marked array
		if oid in self.transaction["up"]:
			if not self.executed:
				self.executed = True
			self.transaction["up"] = []
		if oid in self.transaction["down"]:
			if not self.executed:
				self.executed = True
			self.transaction["down"] = []
		#
		###
		if oid in self.marked:
			if not self.executed:
				self.executed = True
			self.marked.remove(oid)

	def tree(self,creds,prefix = "", oid = None, root = False):
		if self.module.mask & flags.Hidden:
			return ""

		if oid:
			if\
				(not oid in self.transaction["up"]) and\
				(not oid in self.transaction["down"]) and\
				(not oid in self.marked):
				return ""
		else:
			if not self.executed:
				return ""
		cancel = ""
		if oid in self.transaction["down"]:
			cancel = "no "

		if not root:
			r = self.bus.put2("orb",{
				"map": "checkacl",
				"creds": creds,
				"url": self.module.prefix,
			})

			if not r:
				return ""

			if (self.module.mask & flags.Begin) and (not self.comment):
				self.comment = [" ",]
			text = ""
			for i in self.comment:
				text += "%s! %s\n" % (prefix,i)
			text += "%s%s%s\n" % (prefix,cancel,self.fn)
			inc = "\t"
		else:
			text = ""
			inc = ""

		k = self.children.values()
		k.sort(ssort)
		for i in k:
			text += i.tree(creds,prefix + inc,oid)
		return text

class manager (ManagerThread):
	'''
	New state manager
	'''

	def post(self):
		self.root = state2(fn=self.prefixes[0], cn=self.prefixes[0], bus=self.bus.put2("dispatcher2","connect"))

	def getRoot(self,message = {}):
		return self.root

	def chain(self,message):
		node = message["node"]
		chain = message["chain"]
		creds = message["creds"]
		oid = message["oid"]
		comment = message["comment"]
		esoteric = message["esoteric"]
		cookie = message["cookie"]
		result = []

		for i in xrange(len(chain)):
			last = True
			if i != len(chain) - 1:
				last = False
				## zero invalid opts for non-last nodes
				chain[i][2] = ""
			(node,ret) = node.parse(creds, cookie, chain[i], comment, oid, last, esoteric)
			if not (chain[i][0].mask & flags.Immediate):
				comment = []
			result.extend(ret)

		return (node,result,comment)

	def signal(self,message):
		self.log("debug","got a signal: %s" % (message))
		root = self.getRoot()
		subs = []
		services = []
		executed = []
		fqcn = message["signal"]
		if \
			(type(fqcn) != types.ListType) and\
			(type(fqcn) != types.TupleType):
			fqcn = (fqcn,)


		shared = utils.opts()

		for s in fqcn:
			for i in root.catchers:
				if i[-1] == "*":
					k = i[:-1]
					if s.startswith(k):
						for l in root.catchers[i]:
							try:
								subs.append((l,l.module.signal_keys[s]))
							except:
								subs.append((l,None))
				else:
					if s == i:
						for l in root.catchers[i]:
							try:
								subs.append((l,l.module.signal_keys[s]))
							except:
								subs.append((l,None))

		for (i,k) in subs:
			if (i.declare_counter == 0) and (not i in executed) and (i.pass_filter(message["data"])):
				executed.append(i)
				locs = utils.opts()
				locs["signal"] = message["data"]
				if k:
					# more intelligent signal
					i.mark(SIGNAL_OID,"up")
					i.mark(SIGNAL_OID,"down")
					services.extend(root.commit(None,"debug",SIGNAL_OID,locs,shared, k | cmask.Call | cmask.Signal)["services"])
				else:
					# brute force commit
					services.extend(i.commit(None,"debug",None,locs,shared, cmask.Force | cmask.Call | cmask.Signal)["services"])

		if services:
			self.bus.put(
				"runner",
				{
					"map": "perform",
					"pipe": None,
					"ctl": None,
					"services": services,
					"flags": emask.Background,
				}
			)

